(function (factory) {
    if (typeof module === 'object' && typeof module.exports === 'object') {
        var v = factory(require, exports); if (v !== undefined) module.exports = v;
    }
    else if (typeof define === 'function' && define.amd) {
        define(["require", "exports", 'dojo-shim/array', 'dojo-shim/Promise', 'dojo-shim/Set', 'dojo-shim/Map', 'dojo-dom/dom', 'dojo-widgets/projector', './_resolveListenersMap'], factory);
    }
})(function (require, exports) {
    "use strict";
    var array_1 = require('dojo-shim/array');
    var Promise_1 = require('dojo-shim/Promise');
    var Set_1 = require('dojo-shim/Set');
    var Map_1 = require('dojo-shim/Map');
    var dom_1 = require('dojo-dom/dom');
    var projector_1 = require('dojo-widgets/projector');
    var _resolveListenersMap_1 = require('./_resolveListenersMap');
    var reservedNames = new Set_1.default([
        // According to <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
        'annotation-xml',
        'color-profile',
        'font-face',
        'font-face-src',
        'font-face-uri',
        'font-face-format',
        'font-face-name',
        'missing-glyph',
        // These are reserved by this module.
        'widget-instance',
        'widget-projector'
    ]);
    // According to <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
    function isValidName(name) {
        if (!/^[a-z]/.test(name)) {
            return false;
        }
        if (name.indexOf('-') === -1) {
            return false;
        }
        if (/[A-Z]/.test(name)) {
            return false;
        }
        if (reservedNames.has(name)) {
            return false;
        }
        // Assume name does not contain other invalid characters.
        // TODO: Are the above rules sufficiently exclusive given the allowed PCENChar characters in the
        // <https://www.w3.org/TR/custom-elements/#valid-custom-element-name> specification?
        return true;
    }
    exports.isValidName = isValidName;
    // <https://www.w3.org/TR/custom-elements/#look-up-a-custom-element-definition> doesn't define *how* names
    // are to be compared. Additionally browsers and document modes differ in the case used for Element#tagName
    // values. Lowercasing the name is the most compatible solution. This is also the approach taken by the
    // Web Components polyfill:
    // <https://github.com/webcomponents/webcomponentsjs/blob/251f4afedec0ce649728fa1cf22e4fc16bf2bea5/src/CustomElements/register.js#L93>
    function normalizeName(name) {
        return name.toLowerCase();
    }
    exports.normalizeName = normalizeName;
    function isCustomElement(registry, name) {
        return name === 'widget-projector' || name === 'widget-instance' || registry.hasCustomElementFactory(name);
    }
    function getCustomElementsByWidgetProjector(registry, root) {
        var allElements = array_1.from(root.getElementsByTagName('*'));
        allElements.unshift(root); // Be inclusive!
        var customElements = [];
        for (var _i = 0, allElements_1 = allElements; _i < allElements_1.length; _i++) {
            var element = allElements_1[_i];
            var name_1 = void 0;
            var tagName = normalizeName(element.tagName);
            if (isCustomElement(registry, tagName)) {
                name_1 = tagName;
            }
            else {
                var attrIs = normalizeName(element.getAttribute('is') || '');
                if (attrIs !== '' && isCustomElement(registry, attrIs)) {
                    name_1 = attrIs;
                }
            }
            if (name_1) {
                customElements.push({ children: [], element: element, name: name_1 });
            }
        }
        // A list of trees, reconstructed from the `customElements`.
        var widgetProjectors = [];
        // Inverse stack of the nodes in the current tree. The deepest node is at the start of the list.
        var inverseStack = [];
        var discardFirstNode = function (element) {
            if (inverseStack.length === 0) {
                return false;
            }
            // Return `true` if the top-most element in the stack does *not* contain `element`.
            return !(inverseStack[0].element.compareDocumentPosition(element) & Node.DOCUMENT_POSITION_CONTAINED_BY);
        };
        // `customElements` is a flat list of elements, in document order. Reconstruct a tree structure where each
        // root is assumed to be a widget projector.
        for (var _a = 0, customElements_1 = customElements; _a < customElements_1.length; _a++) {
            var custom = customElements_1[_a];
            // Remove nodes from the stack that do not contain the element.
            while (discardFirstNode(custom.element)) {
                inverseStack.shift();
            }
            // Start a new tree if the element is not contained in any existing node.
            if (inverseStack.length === 0) {
                // Don't costruct an invalid tree.
                if (custom.name !== 'widget-projector') {
                    throw new Error('Custom tags must be rooted in a widget-projector');
                }
                widgetProjectors.push(custom);
            }
            else {
                // Don't costruct an invalid tree.
                if (custom.name === 'widget-projector') {
                    throw new Error('widget-projector cannot contain another widget-projector');
                }
                inverseStack[0].children.push(custom);
            }
            // Prepare for the next iteration.
            inverseStack.unshift(custom);
        }
        return widgetProjectors;
    }
    function getIdFromAttributes(element) {
        return element.getAttribute('data-widget-id') || element.getAttribute('id') || undefined;
    }
    function resolveOptions(registry, element, idFromAttributes) {
        var str = element.getAttribute('data-options') || '';
        if (!str) {
            return idFromAttributes ? { id: idFromAttributes } : null;
        }
        var options;
        try {
            options = JSON.parse(str);
        }
        catch (err) {
            throw new SyntaxError("Invalid data-options: " + err.message + " (in " + JSON.stringify(str) + ")");
        }
        if (!options || typeof options !== 'object') {
            throw new TypeError("Expected object from data-options (in " + JSON.stringify(str) + ")");
        }
        if (!('id' in options) && idFromAttributes) {
            options.id = idFromAttributes;
        }
        var promises = [];
        if ('stateFrom' in options) {
            var stateFrom = options.stateFrom;
            if (!stateFrom || typeof stateFrom !== 'string') {
                throw new TypeError("Expected stateFrom value in data-options to be a non-empty string (in " + JSON.stringify(str) + ")");
            }
            promises.push(registry.getStore(stateFrom).then(function (store) {
                options.stateFrom = store;
            }));
        }
        if ('listeners' in options) {
            var listeners = options.listeners;
            var valid = true;
            if (!listeners || typeof listeners !== 'object') {
                valid = false;
            }
            else {
                // Prefer breaking a labeled loop over nesting Array#some() calls or repeating the throwing of
                // the TypeError.
                check: for (var eventType in listeners) {
                    var value = listeners[eventType];
                    if (Array.isArray(value)) {
                        for (var _i = 0, value_1 = value; _i < value_1.length; _i++) {
                            var identifier = value_1[_i];
                            if (typeof identifier !== 'string') {
                                valid = false;
                                break check;
                            }
                        }
                    }
                    else if (typeof value !== 'string') {
                        valid = false;
                        break check;
                    }
                }
            }
            if (!valid) {
                throw new TypeError("Expected listeners value in data-options to be a widget listeners map with action identifiers (in " + JSON.stringify(str) + ")");
            }
            promises.push(_resolveListenersMap_1.default(registry, listeners).then(function (map) {
                options.listeners = map;
            }));
        }
        if (promises.length > 0) {
            return Promise_1.default.all(promises).then(function () {
                return options;
            });
        }
        return options;
    }
    function resolveStateFromAttribute(registry, element) {
        var stateFrom = element.getAttribute('data-state-from');
        return stateFrom ? registry.getStore(stateFrom) : null;
    }
    function getInitialState(element) {
        var str = element.getAttribute('data-state') || '';
        if (!str) {
            return null;
        }
        var initialState;
        try {
            initialState = JSON.parse(str);
        }
        catch (err) {
            throw new SyntaxError("Invalid data-state: " + err.message + " (in " + JSON.stringify(str) + ")");
        }
        if (!initialState || typeof initialState !== 'object') {
            throw new TypeError("Expected object from data-state (in " + JSON.stringify(str) + ")");
        }
        return initialState;
    }
    var noop = function () { };
    function realizeCustomElements(registry, defaultStore, root) {
        // Bottom up, breadth first queue of custom elements who's children's widgets need to be appended to
        // their own widget. Combined for all widget projectors.
        var appendQueue = [];
        // All identified widgets.
        var identifiedWidgets = new Map_1.default();
        // For each projector, track the immediate custom element descendants. These placeholder
        // elements will be replaced with rendered widgets.
        var immediatePlaceholderLookup = new Map_1.default();
        // Projector instances for each widget projector.
        var projectors = [];
        // Widgets that are created during realization (not registered instances).
        var managedWidgets = [];
        // Return a new promise here so API errors can be thrown in the executor, while still resulting in a
        // promise rejection.
        return new Promise_1.default(function (resolve) {
            // Flat list of all widgets that are being loaded.
            var loadedWidgets = [];
            var widgetProjectors = getCustomElementsByWidgetProjector(registry, root);
            for (var _i = 0, widgetProjectors_1 = widgetProjectors; _i < widgetProjectors_1.length; _i++) {
                var _a = widgetProjectors_1[_i], children = _a.children, root_1 = _a.element;
                var projector = projector_1.createProjector({ root: root_1 });
                immediatePlaceholderLookup.set(projector, children);
                projectors.push(projector);
                var projectorStateFrom = resolveStateFromAttribute(registry, root_1);
                // Recursion-free, depth first processing of the tree.
                var processing = [children];
                while (processing.length > 0) {
                    var _loop_1 = function(custom) {
                        var isWidgetInstance = custom.name === 'widget-instance';
                        var id = getIdFromAttributes(custom.element);
                        var promise = null;
                        if (isWidgetInstance) {
                            if (!id) {
                                throw new Error('Cannot resolve widget for a custom element without \'data-widget-id\' or \'id\' attributes');
                            }
                            promise = registry.getWidget(id);
                        }
                        else {
                            promise = Promise_1.default.all([
                                registry.getCustomElementFactory(custom.name),
                                resolveOptions(registry, custom.element, id),
                                resolveStateFromAttribute(registry, custom.element),
                                projectorStateFrom
                            ]).then(function (_a) {
                                var _factory = _a[0], _options = _a[1], _store = _a[2], projectorStore = _a[3];
                                var factory = _factory;
                                var options = _options || {};
                                // `data-state-from` store of the element takes precedence, then of the projector, then
                                // the application's default store.
                                var store = _store || projectorStore || defaultStore;
                                id = options.id;
                                // If the widget has an ID, but stateFrom was not in its `data-options` attribute, and
                                // either its `data-state-from` attribute resolved to a store, or there is a default
                                // store, set the stateFrom option to the `data-state-from` or default store.
                                if (id && !options.stateFrom && store) {
                                    options.stateFrom = store;
                                }
                                if (id && options.stateFrom) {
                                    var initialState = getInitialState(custom.element);
                                    if (initialState) {
                                        return options.stateFrom.patch(initialState, { id: id }).then(function () {
                                            return factory(options);
                                        });
                                    }
                                }
                                return factory(options);
                            });
                        }
                        loadedWidgets.push(promise.then(function (widget) {
                            // Ensure identified widgets are unique.
                            if (id !== undefined) {
                                if (identifiedWidgets.has(id)) {
                                    throw new Error("A widget with ID '" + id + "' already exists");
                                }
                                identifiedWidgets.set(id, widget);
                            }
                            // Store the widget for easy access.
                            custom.widget = widget;
                            // Widget instances come straight from the registry, but the other widgets were created
                            // whilst realizing the custom elements. These should be managed.
                            if (!isWidgetInstance) {
                                managedWidgets.push(widget);
                            }
                            return widget;
                        }));
                        if (custom.children.length > 0) {
                            // Ensure the children are processed.
                            processing.push(custom.children);
                            // Ensure the children are appended to their parent.
                            appendQueue.unshift(custom);
                        }
                    };
                    for (var _b = 0, _c = processing.shift(); _b < _c.length; _b++) {
                        var custom = _c[_b];
                        _loop_1(custom);
                    }
                }
            }
            // Wait for all widgets to be loaded in parallel.
            resolve(Promise_1.default.all(loadedWidgets));
        }).then(function (widgets) {
            // Guard against improper widget usage.
            var uniques = new Set_1.default(widgets);
            if (uniques.size !== widgets.length) {
                throw new Error('Cannot attach a widget multiple times');
            }
            for (var _i = 0, widgets_1 = widgets; _i < widgets_1.length; _i++) {
                var widget = widgets_1[_i];
                // <any> hammer because `widget` could be anything.
                if (widget.parent) {
                    throw new Error('Cannot attach a widget that already has a parent');
                }
            }
            // Build up the widget hierarchy.
            for (var _a = 0, appendQueue_1 = appendQueue; _a < appendQueue_1.length; _a++) {
                var custom = appendQueue_1[_a];
                // Assume the widget has the ParentListMixin.
                var parent_1 = custom.widget;
                var widgets_2 = custom.children.map(function (child) { return child.widget; });
                parent_1.append(widgets_2);
            }
            // Attach all projectors at the same time.
            var attachedProjectors = projectors.map(function (projector) {
                var immediatePlaceholders = immediatePlaceholderLookup.get(projector);
                immediatePlaceholderLookup.delete(projector);
                // Append the top-level widgets to the projector.
                projector.append(immediatePlaceholders.map(function (custom) { return custom.widget; }));
                // Get ready to replace the placeholder elements as soon as the widgets have rendered.
                var handle = projector.on('attach', function () {
                    handle.destroy();
                    var root = projector.root;
                    // Rendered widgets start at this offset.
                    var offset = root.childNodes.length - immediatePlaceholders.length;
                    for (var _i = 0, immediatePlaceholders_1 = immediatePlaceholders; _i < immediatePlaceholders_1.length; _i++) {
                        var placeholder = immediatePlaceholders_1[_i].element;
                        dom_1.place(root.childNodes[offset], dom_1.Position.Replace, placeholder);
                    }
                });
                // Now attach the projector.
                return projector.attach({ type: 'merge' });
            });
            // Wait for the projectors to be attached.
            return Promise_1.default.all(attachedProjectors);
        }).then(function () {
            return {
                destroy: function () {
                    this.destroy = noop;
                    for (var _i = 0, projectors_1 = projectors; _i < projectors_1.length; _i++) {
                        var p = projectors_1[_i];
                        p.destroy();
                    }
                    for (var _a = 0, managedWidgets_1 = managedWidgets; _a < managedWidgets_1.length; _a++) {
                        var w = managedWidgets_1[_a];
                        w.destroy();
                    }
                },
                getWidget: function (id) {
                    return identifiedWidgets.get(id) || null;
                }
            };
        });
    }
    Object.defineProperty(exports, "__esModule", { value: true });
    exports.default = realizeCustomElements;
    ;
});
//# sourceMappingURL=_debug/_realizeCustomElements.js.map