(function (factory) {
    if (typeof module === 'object' && typeof module.exports === 'object') {
        var v = factory(require, exports); if (v !== undefined) module.exports = v;
    }
    else if (typeof define === 'function' && define.amd) {
        define(["require", "exports", 'dojo-compose/compose', 'dojo-shim/Promise', 'dojo-shim/WeakMap', './IdentityRegistry', './_factories', './_moduleResolver', './_realizeCustomElements'], factory);
    }
})(function (require, exports) {
    "use strict";
    var compose_1 = require('dojo-compose/compose');
    var Promise_1 = require('dojo-shim/Promise');
    var WeakMap_1 = require('dojo-shim/WeakMap');
    var IdentityRegistry_1 = require('./IdentityRegistry');
    var _factories_1 = require('./_factories');
    var _moduleResolver_1 = require('./_moduleResolver');
    var _realizeCustomElements_1 = require('./_realizeCustomElements');
    var noop = function () { };
    var actions = new WeakMap_1.default();
    var customElements = new WeakMap_1.default();
    var stores = new WeakMap_1.default();
    var widgets = new WeakMap_1.default();
    var createApp = compose_1.default({
        registerAction: function (id, action) {
            var _this = this;
            var promise = new Promise_1.default(function (resolve) {
                resolve(action.configure(_this._registry));
            }).then(function () { return action; });
            var registryHandle = actions.get(this).register(id, function () { return promise; });
            return {
                destroy: function () {
                    this.destroy = noop;
                    registryHandle.destroy();
                }
            };
        },
        registerActionFactory: function (id, factory) {
            var _this = this;
            var registryHandle = actions.get(this).register(id, function () {
                var promise = Promise_1.default.resolve().then(function () {
                    // Always call the factory in a future turn. This harmonizes behavior regardless of whether the
                    // factory is registered through this method or loaded from a definition.
                    return factory(_this._registry);
                });
                // Replace the registered factory to ensure next time this action is needed, the same action is returned.
                registryHandle.destroy();
                registryHandle = actions.get(_this).register(id, function () { return promise; });
                return promise.then(function (action) {
                    return Promise_1.default.resolve(action.configure(_this._registry)).then(function () { return action; });
                });
            });
            return {
                destroy: function () {
                    this.destroy = noop;
                    registryHandle.destroy();
                }
            };
        },
        registerCustomElementFactory: function (name, factory) {
            if (!_realizeCustomElements_1.isValidName(name)) {
                throw new SyntaxError("'" + name + "' is not a valid custom element name'");
            }
            // Wrap the factory since the registry cannot store frozen factories, and dojo-compose creates
            // frozen factories…
            var wrapped = function (options) { return factory(options); };
            // Note that each custom element requires a new widget, so there's no need to replace the
            // registered factory.
            var registryHandle = customElements.get(this).register(_realizeCustomElements_1.normalizeName(name), wrapped);
            return {
                destroy: function () {
                    this.destroy = noop;
                    registryHandle.destroy();
                }
            };
        },
        registerStore: function (id, store) {
            var promise = Promise_1.default.resolve(store);
            return stores.get(this).register(id, function () { return promise; });
        },
        registerStoreFactory: function (id, factory) {
            var _this = this;
            var registryHandle = stores.get(this).register(id, function () {
                var promise = Promise_1.default.resolve().then(function () {
                    // Always call the factory in a future turn. This harmonizes behavior regardless of whether the
                    // factory is registered through this method or loaded from a definition.
                    return factory();
                });
                // Replace the registered factory to ensure next time this store is needed, the same store is returned.
                registryHandle.destroy();
                registryHandle = stores.get(_this).register(id, function () { return promise; });
                return promise;
            });
            return {
                destroy: function () {
                    this.destroy = noop;
                    registryHandle.destroy();
                }
            };
        },
        registerWidget: function (id, widget) {
            var promise = Promise_1.default.resolve(widget);
            return widgets.get(this).register(id, function () { return promise; });
        },
        registerWidgetFactory: function (id, factory) {
            var _this = this;
            var registryHandle = widgets.get(this).register(id, function () {
                var promise = Promise_1.default.resolve().then(function () {
                    // Always call the factory in a future turn. This harmonizes behavior regardless of whether the
                    // factory is registered through this method or loaded from a definition.
                    var options = { id: id };
                    if (_this.defaultStore) {
                        options.stateFrom = _this.defaultStore;
                    }
                    return factory(options);
                });
                // Replace the registered factory to ensure next time this widget is needed, the same widget is returned.
                registryHandle.destroy();
                registryHandle = widgets.get(_this).register(id, function () { return promise; });
                return promise;
            });
            return {
                destroy: function () {
                    this.destroy = noop;
                    registryHandle.destroy();
                }
            };
        },
        loadDefinition: function (_a) {
            var actions = _a.actions, customElements = _a.customElements, stores = _a.stores, widgets = _a.widgets;
            var app = this;
            var handles = [];
            if (actions) {
                for (var _i = 0, actions_1 = actions; _i < actions_1.length; _i++) {
                    var definition = actions_1[_i];
                    var factory = _factories_1.makeActionFactory(definition, app._resolveMid);
                    var handle = app.registerActionFactory(definition.id, factory);
                    handles.push(handle);
                }
            }
            if (customElements) {
                for (var _b = 0, customElements_1 = customElements; _b < customElements_1.length; _b++) {
                    var definition = customElements_1[_b];
                    var factory = _factories_1.makeCustomElementFactory(definition, app._resolveMid);
                    var handle = app.registerCustomElementFactory(definition.name, factory);
                    handles.push(handle);
                }
            }
            if (stores) {
                for (var _c = 0, stores_1 = stores; _c < stores_1.length; _c++) {
                    var definition = stores_1[_c];
                    var factory = _factories_1.makeStoreFactory(definition, app._resolveMid);
                    var handle = app.registerStoreFactory(definition.id, factory);
                    handles.push(handle);
                }
            }
            if (widgets) {
                for (var _d = 0, widgets_1 = widgets; _d < widgets_1.length; _d++) {
                    var definition = widgets_1[_d];
                    var factory = _factories_1.makeWidgetFactory(definition, app._resolveMid, app);
                    var handle = app.registerWidgetFactory(definition.id, factory);
                    handles.push(handle);
                }
            }
            return {
                destroy: function () {
                    for (var _i = 0, _a = handles.splice(0, handles.length); _i < _a.length; _i++) {
                        var handle = _a[_i];
                        handle.destroy();
                    }
                }
            };
        },
        realize: function (root) {
            return _realizeCustomElements_1.default(this._registry, this.defaultStore, root);
        }
    })
        .mixin({
        mixin: {
            getAction: function (id) {
                var _this = this;
                return new Promise_1.default(function (resolve) {
                    resolve(actions.get(_this).get(id)());
                });
            },
            hasAction: function (id) {
                return actions.get(this).hasId(id);
            },
            getCustomElementFactory: function (name) {
                return customElements.get(this).get(name);
            },
            hasCustomElementFactory: function (name) {
                return customElements.get(this).hasId(name);
            },
            getStore: function (id) {
                var _this = this;
                return new Promise_1.default(function (resolve) {
                    resolve(stores.get(_this).get(id)());
                });
            },
            hasStore: function (id) {
                return stores.get(this).hasId(id);
            },
            getWidget: function (id) {
                var _this = this;
                return new Promise_1.default(function (resolve) {
                    resolve(widgets.get(_this).get(id)());
                });
            },
            hasWidget: function (id) {
                return widgets.get(this).hasId(id);
            }
        },
        initialize: function (instance, _a) {
            var _b = _a === void 0 ? {} : _a, _c = _b.defaultStore, defaultStore = _c === void 0 ? null : _c, _d = _b.toAbsMid, toAbsMid = _d === void 0 ? function (moduleId) { return moduleId; } : _d;
            instance._registry = {
                getAction: instance.getAction.bind(instance),
                hasAction: instance.hasAction.bind(instance),
                getCustomElementFactory: instance.getCustomElementFactory.bind(instance),
                hasCustomElementFactory: instance.hasCustomElementFactory.bind(instance),
                getStore: instance.getStore.bind(instance),
                hasStore: instance.hasStore.bind(instance),
                getWidget: instance.getWidget.bind(instance),
                hasWidget: instance.hasWidget.bind(instance)
            };
            Object.freeze(instance._registry);
            Object.defineProperty(instance, 'defaultStore', {
                configurable: false,
                enumerable: true,
                value: defaultStore,
                writable: false
            });
            instance._resolveMid = _moduleResolver_1.default(toAbsMid);
            actions.set(instance, new IdentityRegistry_1.default());
            customElements.set(instance, new IdentityRegistry_1.default());
            stores.set(instance, new IdentityRegistry_1.default());
            widgets.set(instance, new IdentityRegistry_1.default());
        }
    });
    Object.defineProperty(exports, "__esModule", { value: true });
    exports.default = createApp;
});
//# sourceMappingURL=_debug/createApp.js.map